import time
import numpy as np
import mxnet as mx

__all__ = ['SplitSampler', 'SampledDataset', 'get_split_samplers']

SPLIT_SEED = int(time.time())

def get_split_samplers(train_dataset, split_ratio=0.8):
    num_samples = len(train_dataset)
    split_idx = int(num_samples * split_ratio)
    # numpy seed for consistency
    indices = list(range(num_samples))
    np.random.seed(SPLIT_SEED)
    np.random.shuffle(indices)
    train_sampler = SplitSampler(indices[0: split_idx])
    val_sampler = SplitSampler(indices[split_idx:num_samples])
    return train_sampler, val_sampler

class SplitSampler(object):
    """Samples elements from [start, start+length) randomly without replacement.

    Parameters
    ----------
    length : int
        Length of the sequence.
    """
    def __init__(self, indices):
        self.indices = indices

    def __iter__(self):
        indices = self.indices
        np.random.shuffle(indices)
        return iter(indices)

    def __len__(self):
        return len(self.indices)


class SampledDataset(mx.gluon.data.Dataset):
    """Dataset with elements chosen by a sampler"""
    def __init__(self, dataset, sampler):
        self._dataset = dataset
        self._sampler = sampler
        self._indices = list(iter(sampler))

    def __len__(self):
        return len(self._sampler)

    def __getitem__(self, idx):
        return self._dataset[self._indices[idx]]



import logging
import multiprocessing
import os
from collections import defaultdict
from datetime import datetime

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from sklearn.model_selection import KFold, StratifiedKFold, RepeatedKFold, RepeatedStratifiedKFold, train_test_split

from .constants import BINARY, REGRESSION, MULTICLASS, SOFTCLASS
from ..metrics import accuracy, root_mean_squared_error, Scorer

logger = logging.getLogger(__name__)


def get_pred_from_proba(y_pred_proba, problem_type=BINARY):
    if problem_type == BINARY:
        y_pred = [1 if pred >= 0.5 else 0 for pred in y_pred_proba]
    elif problem_type == REGRESSION:
        y_pred = y_pred_proba
    else:
        y_pred = np.argmax(y_pred_proba, axis=1)
    return y_pred


def generate_kfold(X, y=None, n_splits=5, random_state=0, stratified=False, n_repeats=1):
    if stratified and (y is not None):
        if n_repeats > 1:
            kf = RepeatedStratifiedKFold(n_splits=n_splits, n_repeats=n_repeats, random_state=random_state)
        else:
            kf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)

        kf.get_n_splits(X, y)
        return [[train_index, test_index] for train_index, test_index in kf.split(X, y)]
    else:
        if n_repeats > 1:
            kf = RepeatedKFold(n_splits=n_splits, n_repeats=n_repeats, random_state=random_state)
        else:
            kf = KFold(n_splits=n_splits, shuffle=True, random_state=random_state)

        kf.get_n_splits(X)
        return [[train_index, test_index] for train_index, test_index in kf.split(X)]


def generate_train_test_split(X: DataFrame, y: Series, problem_type: str, test_size: float = 0.1, random_state=0) -> (DataFrame, DataFrame, Series, Series):
    if (test_size <= 0.0) or (test_size >= 1.0):
        raise ValueError("fraction of data to hold-out must be specified between 0 and 1")

    if problem_type in [REGRESSION, SOFTCLASS]:
        stratify = None
    else:
        stratify = y

    # TODO: Enable stratified split when y class would result in 0 samples in test.
    #  One approach: extract low frequency classes from X/y, add back (1-test_size)% to X_train, y_train, rest to X_test
    #  Essentially stratify the high frequency classes, random the low frequency (While ensuring at least 1 example stays for each low frequency in train!)
    #  Alternatively, don't test low frequency at all, trust it to work in train set. Risky, but highest quality for predictions.
    X_train, X_test, y_train, y_test = train_test_split(X, y.values, test_size=test_size, shuffle=True, random_state=random_state, stratify=stratify)
    if problem_type != SOFTCLASS:
        y_train = pd.Series(y_train, index=X_train.index)
        y_test = pd.Series(y_test, index=X_test.index)
    else:
        y_train = pd.DataFrame(y_train, index=X_train.index)
        y_test = pd.DataFrame(y_test, index=X_test.index)
    return X_train, X_test, y_train, y_test


def setup_outputdir(output_directory):
    if output_directory is None:
        utcnow = datetime.utcnow()
        timestamp = utcnow.strftime("%Y%m%d_%H%M%S")
        output_directory = f"AutogluonModels/ag-{timestamp}{os.path.sep}"
        for i in range(1, 1000):
            try:
                os.makedirs(output_directory, exist_ok=False)
                break
            except FileExistsError as e:
                output_directory = f"AutogluonModels/ag-{timestamp}-{i:03d}{os.path.sep}"
        else:
            raise RuntimeError("more than 1000 jobs launched in the same second")
        logger.log(25, f"No output_directory specified. Models will be saved in: {output_directory}")
    output_directory = os.path.expanduser(output_directory)  # replace ~ with absolute path if it exists
    if output_directory[-1] != os.path.sep:
        output_directory = output_directory + os.path.sep
    return output_directory


def setup_compute(nthreads_per_trial, ngpus_per_trial):
    if nthreads_per_trial is None:
        nthreads_per_trial = multiprocessing.cpu_count()  # Use all of processing power / trial by default. To use just half: # int(np.floor(multiprocessing.cpu_count()/2))

    if ngpus_per_trial is None:
        ngpus_per_trial = 0  # do not use GPU by default
    elif ngpus_per_trial > 1:
        ngpus_per_trial = 1
        logger.debug("tabular_prediction currently doesn't use >1 GPU per training run. ngpus_per_trial set = 1")
    return nthreads_per_trial, ngpus_per_trial


def setup_trial_limits(time_limits, num_trials, hyperparameters):
    """ Adjust default time limits / num_trials """
    if num_trials is None:
        if time_limits is None:
            time_limits = 10 * 60  # run for 10min by default
        time_limits /= float(len(hyperparameters))  # each model type gets half the available time
        num_trials = 1000  # run up to 1000 trials (or as you can within the given time_limits)
    elif time_limits is None:
        time_limits = int(1e6)  # user only specified num_trials, so run all of them regardless of time-limits
    else:
        time_limits /= float(len(hyperparameters))  # each model type gets half the available time

    if time_limits <= 10:  # threshold = 10sec, ie. too little time to run >1 trial.
        num_trials = 1
    time_limits *= 0.9  # reduce slightly to account for extra time overhead
    return time_limits, num_trials


def dd_list():
    return defaultdict(list)


def get_leaderboard_pareto_frontier(leaderboard: DataFrame, score_col='score_val', inference_time_col='pred_time_val_full') -> DataFrame:
    """
    Given a set of models, returns in ranked order from best score to worst score models which satisfy the criteria:
    1. No other model in the set has both a lower inference time and a better or equal score.

    :param leaderboard: Leaderboard DataFrame of model info containing score_col and inference_time_col
    :param score_col: Column name in leaderboard of model score values
    :param inference_time_col: Column name in leaderboard of model inference times
    :return: Subset of the original leaderboard DataFrame containing only models that are a valid optimal choice at different valuations of score and inference time.
    """
    leaderboard = leaderboard.sort_values(by=[score_col, inference_time_col], ascending=[False, True]).reset_index(drop=True)
    leaderboard_unique = leaderboard.drop_duplicates(subset=[score_col])

    pareto_frontier = []
    inference_time_min = None
    for index, row in leaderboard_unique.iterrows():
        if row[inference_time_col] is None or row[score_col] is None:
            pass
        elif (inference_time_min is None) or (row[inference_time_col] < inference_time_min):
            inference_time_min = row[inference_time_col]
            pareto_frontier.append(index)
    leaderboard_pareto_frontier = leaderboard_unique.loc[pareto_frontier].reset_index(drop=True)
    return leaderboard_pareto_frontier


def shuffle_df_rows(X: DataFrame, seed=0, reset_index=True):
    """Returns DataFrame with rows shuffled based on seed value."""
    row_count = X.shape[0]
    np.random.seed(seed)
    rand_shuffle = np.random.randint(0, row_count, size=row_count)
    X_shuffled = X.iloc[rand_shuffle]
    if reset_index:
        X_shuffled.reset_index(inplace=True, drop=True)
    return X_shuffled


def normalize_pred_probas(y_predprob, problem_type, eps=1e-7):
    """ Remaps the predicted probabilities to ensure there are no zeros (needed for certain metrics like log-loss)
        and that no predicted probability exceeds [0,1] (eg. in distillation when classification is treated as regression).
        Args:
            y_predprob: 1D (for binary classification) or 2D (for multiclass) numpy array of predicted probabilities
            problem_type: We only consider normalization if the problem_type is one of: [BINARY, MULTICLASS, SOFTCLASS]
            eps: controls around how far from 0 remapped predicted probabilities should be (larger `eps` means predicted probabilities will lie further from 0).
    """
    if (problem_type == REGRESSION) and (len(y_predprob.shape) > 1) and (y_predprob.shape[1] > 1):
        problem_type = SOFTCLASS  # this was MULTICLASS problem converted to REGRESSION (as done in distillation)

    if problem_type in [BINARY, REGRESSION]:
        if len(y_predprob.shape) > 1 and min(y_predprob.shape) > 1:
            raise ValueError(f"cannot call normalize_pred_probas with problem_type={problem_type} and y_predprob.shape=={y_predprob.shape}")
        return normalize_binary_probas(y_predprob, eps)
    elif problem_type in [MULTICLASS, SOFTCLASS]:  # clip all probs below at eps and then renormalize
        if len(y_predprob.shape) == 1:
            return normalize_binary_probas(y_predprob, eps)
        else:
            return normalize_multi_probas(y_predprob, eps)
    else:
        raise ValueError(f"Invalid problem_type")


def normalize_binary_probas(y_predprob, eps):
    """ Remaps the predicted probabilities to open interval (0,1) while maintaining rank order """
    (pmin,pmax) = (eps, 1-eps)  # predicted probs outside this range will be remapped into (0,1)
    which_toobig = y_predprob > pmax
    if np.sum(which_toobig) > 0:  # remap overly large probs
        y_predprob = np.logical_not(which_toobig)*y_predprob + which_toobig*(1-(eps*np.exp(-(y_predprob-pmax))))
    which_toosmall = y_predprob < pmin
    if np.sum(which_toosmall) > 0:  # remap overly small probs
        y_predprob = np.logical_not(which_toosmall)*y_predprob + which_toosmall*eps*np.exp(-(pmin-y_predprob))
    return y_predprob


def normalize_multi_probas(y_predprob, eps):
    """ Remaps the predicted probabilities to lie in (0,1) where eps controls how far from 0 smallest class-probability lies """
    min_predprob = np.min(y_predprob)
    if min_predprob < 0:  # ensure nonnegative rows
        most_negative_rowvals = np.clip(np.min(y_predprob, axis=1), a_min=None, a_max=0)
        y_predprob = y_predprob - most_negative_rowvals[:,None]
    if min_predprob < eps:
        y_predprob = np.clip(y_predprob, a_min=eps, a_max=None)  # ensure no entries < eps
        y_predprob = y_predprob / y_predprob.sum(axis=1, keepdims=1)  # renormalize
    return y_predprob


def infer_problem_type(y: Series):
    """ Identifies which type of prediction problem we are interested in (if user has not specified).
        Ie. binary classification, multi-class classification, or regression.
    """
    if len(y) == 0:
        raise ValueError("provided labels cannot have length = 0")
    y = y.dropna()  # Remove missing values from y (there should not be any though as they were removed in Learner.general_data_processing())
    num_rows = len(y)

    unique_values = y.unique()

    MULTICLASS_LIMIT = 1000  # if numeric and class count would be above this amount, assume it is regression
    if num_rows > 1000:
        REGRESS_THRESHOLD = 0.05  # if the unique-ratio is less than this, we assume multiclass classification, even when labels are integers
    else:
        REGRESS_THRESHOLD = 0.1

    unique_count = len(unique_values)
    if unique_count == 2:
        problem_type = BINARY
        reason = "only two unique label-values observed"
    elif y.dtype.name in ['object', 'category']:
        problem_type = MULTICLASS
        reason = f"dtype of label-column == {y.dtype.name}"
    elif np.issubdtype(y.dtype, np.floating):
        unique_ratio = unique_count / float(num_rows)
        if (unique_ratio <= REGRESS_THRESHOLD) and (unique_count <= MULTICLASS_LIMIT):
            try:
                can_convert_to_int = np.array_equal(y, y.astype(int))
                if can_convert_to_int:
                    problem_type = MULTICLASS
                    reason = "dtype of label-column == float, but few unique label-values observed and label-values can be converted to int"
                else:
                    problem_type = REGRESSION
                    reason = "dtype of label-column == float and label-values can't be converted to int"
            except:
                problem_type = REGRESSION
                reason = "dtype of label-column == float and label-values can't be converted to int"
        else:
            problem_type = REGRESSION
            reason = "dtype of label-column == float and many unique label-values observed"
    elif np.issubdtype(y.dtype, np.integer):
        unique_ratio = unique_count / float(num_rows)
        if (unique_ratio <= REGRESS_THRESHOLD) and (unique_count <= MULTICLASS_LIMIT):
            problem_type = MULTICLASS  # TODO: Check if integers are from 0 to n-1 for n unique values, if they have a wide spread, it could still be regression
            reason = "dtype of label-column == int, but few unique label-values observed"
        else:
            problem_type = REGRESSION
            reason = "dtype of label-column == int and many unique label-values observed"
    else:
        raise NotImplementedError(f'label dtype {y.dtype} not supported!')
    logger.log(25, f"AutoGluon infers your prediction problem is: '{problem_type}' (because {reason}).")

    # TODO: Move this outside of this function so it is visible even if problem type was not inferred.
    if problem_type in [BINARY, MULTICLASS]:
        if unique_count > 10:
            logger.log(20, f'\tFirst 10 (of {unique_count}) unique label values:  {list(unique_values[:10])}')
        else:
            logger.log(20, f'\t{unique_count} unique label values:  {list(unique_values)}')
    elif problem_type == REGRESSION:
        y_max = y.max()
        y_min = y.min()
        y_mean = y.mean()
        y_stddev = y.std()
        logger.log(20, f'\tLabel info (max, min, mean, stddev): ({y_max}, {y_min}, {round(y_mean, 5)}, {round(y_stddev, 5)})')

    logger.log(25, f"\tIf '{problem_type}' is not the correct problem_type, please manually specify the problem_type argument in fit() (You may specify problem_type as one of: {[BINARY, MULTICLASS, REGRESSION]})")
    return problem_type


def infer_eval_metric(problem_type: str) -> Scorer:
    """Infers appropriate default eval metric based on problem_type. Useful when no eval_metric was provided."""
    if problem_type == BINARY:
        return accuracy
    elif problem_type == MULTICLASS:
        return accuracy
    else:
        return root_mean_squared_error


def default_holdout_frac(num_train_rows, hyperparameter_tune=False):
    """ Returns default holdout_frac used in fit().
        Between row count 5,000 and 25,000 keep 0.1 holdout_frac, as we want to grow validation set to a stable 2500 examples.
    """
    if num_train_rows < 5000:
        holdout_frac = max(0.1, min(0.2, 500.0 / num_train_rows))
    else:
        holdout_frac = max(0.01, min(0.1, 2500.0 / num_train_rows))

    if hyperparameter_tune:
        holdout_frac = min(0.2, holdout_frac * 2)  # We want to allocate more validation data for HPO to avoid overfitting

    return holdout_frac


def augment_rare_classes(X, label, threshold):
    """ Use this method when using certain eval_metrics like log_loss, for which no classes may be filtered out.
        This method will augment dataset with additional examples of rare classes.
    """
    class_counts = X[label].value_counts()
    class_counts_invalid = class_counts[class_counts < threshold]
    if len(class_counts_invalid) == 0:
        logger.debug("augment_rare_classes did not need to duplicate any data from rare classes")
        return X

    aug_df = None
    for clss, n_clss in class_counts_invalid.iteritems():
        n_toadd = threshold - n_clss
        clss_df = X.loc[X[label] == clss]
        if aug_df is None:
            aug_df = clss_df[:0].copy()
        duplicate_times = int(np.floor(n_toadd / n_clss))
        remainder = n_toadd % n_clss
        new_df = clss_df.copy()
        new_df = new_df[:remainder]
        while duplicate_times > 0:
            logger.debug(f"Duplicating data from rare class: {clss}")
            duplicate_times -= 1
            new_df = new_df.append(clss_df.copy())
        aug_df = aug_df.append(new_df.copy())

    X = X.append(aug_df)
    class_counts = X[label].value_counts()
    class_counts_invalid = class_counts[class_counts < threshold]
    if len(class_counts_invalid) > 0:
        raise RuntimeError("augment_rare_classes failed to produce enough data from rare classes")
    logger.log(15, "Replicated some data from rare classes in training set because eval_metric requires all classes")
    return X



from typing import Tuple, Optional
import mxnet as mx

from .kernel import KernelFunction
from .mean import MeanFunction
from .posterior_utils import Tensor, mxnet_F, cholesky_computations, predict_posterior_marginals, sample_posterior_marginals, sample_posterior_joint, cholesky_update, negative_log_marginal_likelihood, mxnet_is_ndarray


class GaussProcPosteriorState(object):
    """
    Represent posterior state for Gaussian process regression model.
    Note that members are immutable. If the posterior state is to be
    updated, a new object is created and returned.

    """
    def __init__(
            self, features: Tensor, targets: Optional[Tensor],
            mean: MeanFunction, kernel: KernelFunction,
            noise_variance: Tensor, debug_log: bool = False,
            test_intermediates: Optional[dict] = None,
            **kwargs):
        """
        If targets has m > 1 columns, they correspond to fantasy samples.

        If targets is None, this is an internal (copy) constructor, where
        kwargs contains chol_fact, pred_mat.

        :param features: Input points X, shape (n, d)
        :param targets: Targets Y, shape (n, m)
        :param mean: Mean function m(X)
        :param kernel: Kernel function k(X, X')
        :param noise_variance: Noise variance sigsq, shape (1,)
        :param test_intermediates: See cholesky_computations
        """
        F = mxnet_F(features)
        self.F = F
        self.mean = mean
        self.kernel = kernel
        if targets is not None:
            targets = F.reshape(targets, shape=(0, -1))
            chol_fact, pred_mat = cholesky_computations(
                F, features, targets, mean, kernel, noise_variance,
                debug_log=debug_log, test_intermediates=test_intermediates)
            if self.is_ndarray():
                # Make sure the computations are done, and not deferred until
                # later. This is important in order to get profiling right
                # (and there is no advantage in deferring at this point)
                chol_fact.wait_to_read()
                pred_mat.wait_to_read()
            # Make copy, just to be safe:
            self.features = features if F == mx.sym else features.copy()
            self.chol_fact = chol_fact
            self.pred_mat = pred_mat
            self._test_intermediates = test_intermediates
        else:
            # Internal (copy) constructor
            self.features = features
            self.chol_fact = kwargs['chol_fact']
            self.pred_mat = kwargs['pred_mat']

    @property
    def num_data(self):
        self._check_is_ndarray()
        return self.features.shape[0]

    @property
    def num_features(self):
        self._check_is_ndarray()
        return self.features.shape[1]

    @property
    def num_fantasies(self):
        self._check_is_ndarray()
        return self.pred_mat.shape[1]

    def _state_args(self):
        return [self.F, self.features, self.mean, self.kernel, self.chol_fact,
            self.pred_mat]

    def predict(self, test_features: Tensor) -> Tuple[Tensor, Tensor]:
        return predict_posterior_marginals(
            *self._state_args(), test_features,
            test_intermediates=self._test_intermediates)

    def sample_joint(self, test_features: Tensor, num_samples: int=1) -> \
            Tensor:
        return sample_posterior_joint(
            *self._state_args(), test_features, num_samples)

    def sample_marginals(self, test_features: Tensor, num_samples: int=1) -> \
            Tensor:
        return sample_posterior_marginals(
            *self._state_args(), test_features, num_samples)

    def neg_log_likelihood(self) -> Tensor:
        """
        Works only if fantasy samples are not used (single targets vector).

        :return: Negative log (marginal) likelihood
        """
        return negative_log_marginal_likelihood(
            self.F, self.chol_fact, self.pred_mat)

    def is_ndarray(self):
        return mxnet_is_ndarray(self.F)

    def _check_is_ndarray(self):
        assert self.is_ndarray(), \
            "Implemented only for mxnet.ndarray, not for mxnet.symbol"


class IncrementalUpdateGPPosteriorState(GaussProcPosteriorState):
    """
    Extension of GaussProcPosteriorState which allows for incremental
    updating, given that a single data case is appended to the training
    set.

    In order to not mutate members, the update method returns a new
    object.

    """
    def __init__(
            self, features: Tensor, targets: Optional[Tensor],
            mean: MeanFunction, kernel: KernelFunction,
            noise_variance: Tensor, **kwargs):
        super(IncrementalUpdateGPPosteriorState, self).__init__(
            features, targets, mean, kernel, noise_variance, **kwargs)
        # Noise variance is needed here for updates (make copy, to be safe)
        self.noise_variance = noise_variance if self.F == mx.sym \
            else noise_variance.copy()

    def update(self, feature: Tensor, target: Tensor) -> \
            'IncrementalUpdateGPPosteriorState':
        """
        :param feature: Additional input xstar, shape (1, d)
        :param target: Additional target ystar, shape (1, m)
        :return: Posterior state for increased data set
        """
        # Ensure feature, target have one row
        F = self.F
        feature = F.reshape(feature, shape=(1, -1))
        target = F.reshape(target, shape=(1, -1))
        if self.is_ndarray():
            assert feature.shape[1] == self.features.shape[1], \
                "feature.shape[1] = {} != {} = self.features.shape[1]".format(
                    feature.shape[1], self.features.shape[1])
            assert target.shape[1] == self.pred_mat.shape[1], \
                "target.shape[1] = {} != {} = self.pred_mat.shape[1]".format(
                    target.shape[1], self.pred_mat.shape[1])
        chol_fact_new, pred_mat_new = cholesky_update(
            F, self.features, self.chol_fact, self.pred_mat, self.mean,
            self.kernel, self.noise_variance, feature, target)
        features_new = F.concat(self.features, feature, dim=0)
        # Create object by calling internal constructor
        state_new = IncrementalUpdateGPPosteriorState(
            features=features_new,
            targets=None,
            mean=self.mean,
            kernel=self.kernel,
            noise_variance=self.noise_variance,
            chol_fact=chol_fact_new,
            pred_mat=pred_mat_new)
        return state_new

    def expand_fantasies(self, num_fantasies: int) -> \
            'IncrementalUpdateGPPosteriorState':
        """
        If this posterior has been created with a single targets vector,
        shape (n, 1), use this to duplicate this vector m = num_fantasies
        times. Call this method before fantasy targets are appended by
        update.

        :param num_fantasies: Number m of fantasy samples
        :return: New state with targets duplicated m times
        """
        assert num_fantasies > 1
        F = self.F
        if self.is_ndarray():
            assert self.pred_mat.shape[1] == 1, \
                "Method requires posterior without fantasy samples"
        pred_mat_new = F.concat(*([self.pred_mat] * num_fantasies), dim=1)
        state_new = IncrementalUpdateGPPosteriorState(
            features=self.features,
            targets=None,
            mean=self.mean,
            kernel=self.kernel,
            noise_variance=self.noise_variance,
            chol_fact=self.chol_fact,
            pred_mat=pred_mat_new)
        return state_new



"""
optimization_utils
==================

Wrapper of SciPy L-BFGS-B optimizer minimizing objectives given by MXNet
executors. The main issue is to map between parameter dictionaries of
mx.ndarray (executor side) and a single flat np.ndarray (optimizer side).
"""

import mxnet as mx
import numpy as np
from scipy import optimize
import ctypes

__all__ = ['apply_lbfgs',
           'from_executor',
           'apply_lbfgs_with_multiple_starts']


# NOTE: _dtype is the dtype used on the NumPy side (argument, value, and
# gradient of objective.
_dtype = np.float64
default_LBFGS_tol, default_LBFGS_maxiter = 1e-5, 500

N_STARTS = 5
STARTING_POINT_RANDOMIZATION_STD = 1.


# Utility functions for (un-)grouping NDArrays into Numpy arrays.
# Useful e.g. to use optimizers from scipy.optimize to optimize
# functions expressed in MXNet.

def _get_name_to_index(nd_arrays, names):
    name_to_index = {}
    global_position = 0
    for name in names:
        a = nd_arrays[name]
        name_to_index[name] = np.arange(global_position, global_position+a.size)
        global_position += a.size
    return name_to_index


def _zeros_like_nd_list(l, dtype):
    """
    Create a Numpy array with size equal to the
    sum of the sizes of all the NDArrays in the list
    of NDArrays l.
    """
    total_size = np.sum([x.size for x in l])
    return np.zeros(total_size, dtype)


def _copy_into(ndarray, nparray):
    """
    Copy the values from the given ndarray into the given (preallocated) numpy array.
    This can be used to avoid extra memory allocation that ndarray.asnumpy()
    performs.
    """
    assert nparray.size == ndarray.size
    assert nparray.flags.f_contiguous and nparray.flags.behaved
    # NOTE: The copy=False variant of NDArray.astype does not seem to work
    if ndarray.dtype != nparray.dtype:
        ndarray = ndarray.astype(nparray.dtype)
    mx.base.check_call(mx.base._LIB.MXNDArraySyncCopyToCPU(
        ndarray.handle,
        nparray.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_size_t(ndarray.size)))


def _copy_to_numpy_array(l, a):
    """
    Copy values from each NDArray in the list l to the numpy array a (in
    order).
    """
    total_size = np.sum([x.size for x in l])
    assert total_size == a.size
    j = 0
    for x in l:
        # a[j:j+x.size] = x.asnumpy().reshape((x.size,))
        _copy_into(x, a[j:j + x.size])
        j += x.size


def _copy_from_numpy_array(a, l):
    """
    Copy values from sub-arrays of the numpy array a to
    the NDArrays in the list l. The sizes of the sub arrays
    correspond to the sizes of the NDArrays, so that this
    performs a copy in the reverse direction of
    copy_to_numpy_array().

    Entries of l can have different dtype than a.
    """
    total_size = np.sum([x.size for x in l])
    assert total_size == a.size
    j = 0
    for x in l:
        x[:] = a[j:j + x.size].reshape(x.shape)
        j += x.size


def _make_objective(
        exec_func, arg_dict, grad_dict, param_names):
    # list of argument and grad NDArrays corresponding to param_names
    param_nd_arrays = [arg_dict[name] for name in param_names]
    grad_nd_arrays = [grad_dict[name] for name in param_names]
    # Numpy arrays for holding parameters and gradients
    # grad output array for direct copy
    grad_numpy_array = _zeros_like_nd_list(grad_nd_arrays, _dtype)

    # Objective and its gradients
    def objective(a):
        # copy parameter values to arg_dict (NDArrays)
        _copy_from_numpy_array(a, param_nd_arrays)
        # compute objective and gradients
        obj_val = exec_func()
        # Copy gradients from grad_dict (NDArray) into numpy array
        _copy_to_numpy_array(grad_nd_arrays, grad_numpy_array)
        return obj_val, grad_numpy_array

    return objective


def _apply_lbfgs_internal(
        exec_func, arg_dict, grad_dict, param_names, param_numpy_array,
        name_to_index, bounds, **kwargs):
    # Define bounds for L-BFGS, None by default
    param_bounds = np.array([(None, None)] * len(param_numpy_array))
    for name, bound in bounds.items():
        if name in param_names:
            param_bounds[name_to_index[name]] = bound
    # Objective from executor
    mx_objective = _make_objective(
        exec_func, arg_dict, grad_dict, param_names)
    # Run L-BFGS-B
    LBFGS_tol = kwargs.get("tol", default_LBFGS_tol)
    LBFGS_maxiter = kwargs.get("maxiter", default_LBFGS_maxiter)
    LBFGS_callback = kwargs.get("callback", None)
    ret_info = None
    try:
        output = optimize.minimize(mx_objective,
                                   param_numpy_array,
                                   jac=True,
                                   method="L-BFGS-B",
                                   bounds=param_bounds,
                                   tol=LBFGS_tol,
                                   options={"maxiter": LBFGS_maxiter},
                                   callback=LBFGS_callback)
        # NOTE: Be aware that the stopping condition based on tol can terminate
        # with a gradient size which is not small.
        # To control L-BFGS convergence conditions for real, have to instead use
        # something like this:
        #                           tol=None,
        #                           options={
        #                               "maxiter": LBFGS_maxiter,
        #                               "ftol": 1e-6,
        #                               "gtol": 1e-1},
        #                           callback=LBFGS_callback)

        # Write result evaluation point back to arg_dict
        optimized_param_numpy_array = output.x
        param_nd_arrays = [arg_dict[name] for name in param_names]
        _copy_from_numpy_array(optimized_param_numpy_array, param_nd_arrays)
    except Exception as inst:
        ret_info = {
            'type': type(inst),
            'args': inst.args,
            'msg': str(inst)}
    return ret_info


# Utility functions for multiple restarts


class ExecutorDecorator:
    """
    This class is a lightweight decorator around the executor passed to L-BFGS
    It adds the functionality of keeping track of the best objective function
    """

    def __init__(self, exec_func):
        self.best_objective = np.inf
        self._exec_func = exec_func

    def exec_func(self):
        objective = self._exec_func()
        self.best_objective = min(self.best_objective, objective)
        return objective


def _deep_copy_arg_dict(input_arg_dict):
    """
    Make a deep copy of the input arg_dict (dict param_name to mx.nd)
    :param input_arg_dict:
    :return: deep copy of input_arg_dict
    """
    output_arg_dict = {}
    for name, param in input_arg_dict.items():
        output_arg_dict[name] = param.copy()
    return output_arg_dict


def _inplace_arg_dict_randomization(arg_dict, mean_arg_dict, bounds, std=STARTING_POINT_RANDOMIZATION_STD):
    """
    In order to initialize L-BFGS from multiple starting points, this function makes it possible to
    randomize, inplace, an arg_dict (as used by executors to communicate parameters to L-BFGS).
    The randomization is centered around mean_arg_dict, with standard deviation std.

    :param arg_dict: dict param_name to mx.nd (as used in executors). This argument is modified inplace
    :param mean_arg_dict: arg_dict around which the random perturbations occur (dict param_name to mx.nd, as used in executors))
    :param bounds: dict param_name to (lower, upper) bounds, as used in L-BFGS
    :param std: standard deviation according to which the (Gaussian) random perturbations happen
    """

    # We check that arg_dict and mean_arg_dict are compatible
    assert arg_dict.keys() == mean_arg_dict.keys()
    for name, param in arg_dict.items():
        assert param.shape == mean_arg_dict[name].shape
        assert param.dtype == mean_arg_dict[name].dtype
        assert param.context == mean_arg_dict[name].context

    # We apply a sort to make the for loop deterministic (especially with the internal calls to mx.random)
    for name, param in sorted(arg_dict.items()):

        arg_dict[name][:] = mean_arg_dict[name] + mx.random.normal(0.0, std, shape=param.shape, dtype=param.dtype, ctx=param.context)

        lower, upper = bounds[name]
        lower = lower if lower is not None else -np.inf
        upper = upper if upper is not None else np.inf

        # We project back arg_dict[name] within its specified lower and upper bounds
        # (in case of we would have perturbed beyond those bounds)
        arg_dict[name][:] = mx.nd.maximum(lower, mx.nd.minimum(upper, arg_dict[name]))


# === Exported functions ===


def apply_lbfgs(exec_func, arg_dict, grad_dict, bounds, **kwargs):
    """Run SciPy L-BFGS-B on criterion given by MXNet code

    Run SciPy L-BFGS-B in order to minimize criterion given by MXNet code.
    Criterion and gradient are computed by:

        crit_val = exec_func()

    Here, arguments are taken from arg_dict, and gradients are written to
    grad_dict (both are dictionaries with NDArray values). crit_val is a
    Python scalar, not an NDArray.
    Think of arg_dict and grad_dict as args and args_grad arguments of an
    MXNet executor.

    The variables which L-BFGS-B is optimizing over are all those in
    grad_dict whose values are not None. Both arg_dict and grad_dict must
    contain values for these keys, and they must have the same shapes.
    Both arg_dict and grad_dict can have additional entries, but these are
    not modified.

    Initial values are taken from arg_dict, and final values are written
    back there.

    L-BFGS-B allows box constraints [a, b] for any coordinate. Here, None
    stands for -infinity (a) or +infinity (b). The default is (None, None),
    so no constraints. In bounds, box constraints can be specified per
    argument (the constraint applies to all coordinates of the argument).
    Pass {} for no constraints.

    If the criterion function is given by an MXNet executor mx_executor,
    you can call

        apply_bfgs(*from_executor(mx_executor), bounds, ...)

    See from_executor comments for details.

    :param exec_func: Function to compute criterion
    :param arg_dict: See above
    :param grad_dict: See above
    :param bounds: See above
    :return: None, or dict with info about exception caught
    """

    param_names = sorted(
        [name for name, value in grad_dict.items() \
         if value is not None])
    name_to_index = _get_name_to_index(arg_dict, param_names)
    # Construct initial evaluation point (NumPy)
    param_nd_arrays = [arg_dict[name] for name in param_names]
    param_numpy_array = _zeros_like_nd_list(param_nd_arrays, _dtype)
    _copy_to_numpy_array(param_nd_arrays, param_numpy_array)

    return _apply_lbfgs_internal(
        exec_func, arg_dict, grad_dict, param_names, param_numpy_array,
        name_to_index, bounds, return_results=False, **kwargs)


def from_executor(executor):
    """Maps MXNet executor to apply_lbfgs arguments

    apply_lbfgs allows to pass exec_func, arg_dict, grad_dict to specify the
    criterion function, its argument and gradient dictionaries. If your
    criterion is given by an MXNet executor mx_executor, you can call

        apply_bfgs(*from_executor(mx_executor), bounds, ...)

    Here, arg_dict = mx_executor.arg_dict, grad_dict = mx_executor.grad_dict.
    This requires that mx_executors represents a loss function (use
    mx.sym.MakeLoss to be safe).

    :param executor: MXNet executor representing a loss function
    :return: exec_func, arg_dict, grad_dict arguments for apply_lbfgs
    """

    def exec_func():
        executor.forward(is_train=True)
        obj_val = executor.outputs[0].asscalar()
        executor.backward()
        return obj_val

    return exec_func, executor.arg_dict, executor.grad_dict


def apply_lbfgs_with_multiple_starts(
        exec_func, arg_dict, grad_dict, bounds, n_starts=N_STARTS, **kwargs):
    """
    When dealing with non-convex problems (e.g., optimization the marginal
    likelihood), we typically need to start from various starting points. This
    function applies this logic around apply_lbfgs, randomizing the starting
    points around the initial values provided in arg_dict (see below
    "copy_of_initial_arg_dict").

    The first optimization happens exactly at arg_dict, so that the case
    n_starts=1 exactly coincides with the previously used apply_lbfgs.
    Importantly, the communication with the L-BFGS solver happens via arg_dict,
    hence all the operations with respect to arg_dict are inplace.

    We catch exceptions and return ret_infos about these. If none of the
    restarts worked, arg_dict is not modified.

    :param exec_func: see above
    :param arg_dict: see above
    :param grad_dict: see above
    :param bounds: see above
    :param n_starts: Number of times we start an optimization with L-BFGS
        (must be >= 1)
    :return: List ret_infos of length n_starts. Entry is None if optimization
        worked, or otherwise has dict with info about exception caught
    """

    assert n_starts >= 1

    copy_of_initial_arg_dict = _deep_copy_arg_dict(arg_dict)
    best_objective_over_restarts = None
    best_arg_dict_over_restarts = copy_of_initial_arg_dict

    # Loop over restarts
    ret_infos = []
    for iter in range(n_starts):
        if iter > 0:
            _inplace_arg_dict_randomization(
                arg_dict, copy_of_initial_arg_dict, bounds)
        decorator = ExecutorDecorator(exec_func)
        ret_info = apply_lbfgs(
            decorator.exec_func, arg_dict, grad_dict, bounds, **kwargs)
        ret_infos.append(ret_info)
        if ret_info is None and (
                best_objective_over_restarts is None or
                decorator.best_objective < best_objective_over_restarts):
            best_objective_over_restarts = decorator.best_objective
            best_arg_dict_over_restarts = _deep_copy_arg_dict(arg_dict)

    # We copy back the values of the best parameters into arg_dict (again,
    # inplace, as required by the executor)
    for name in arg_dict.keys():
        arg_dict[name][:] = best_arg_dict_over_restarts[name]
    return ret_infos



import json
import collections
import mxnet as mx
from mxnet import gluon
from ...core.space import Categorical, Space, _strip_config_space

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = ['enas_unit', 'enas_net',
           'Zero_Unit', 'ENAS_Unit', 'ENAS_Sequential']

def enas_unit(**kwvars):
    def registered_class(Cls):
        class enas_unit(ENAS_Unit):
            def __init__(self, *args, **kwargs):
                kwvars.update(kwargs)
                with_zero=False
                if 'with_zero' in kwvars:
                    with_zero = kwvars.pop('with_zero')
                blocks = []
                self._args = []
                for arg in self.get_config_grid(kwvars):
                    blocks.append(Cls(*args, **arg))
                    self._args.append(json.dumps(arg))
                if with_zero:
                    self._args.append(None)
                super().__init__(*blocks, with_zero=with_zero)

            @property
            def node(self):
                arg = self._args[self.index]
                if arg is None: return arg
                summary = {}
                name = self.module_list[self.index].__class__.__name__ + '('
                for k, v in json.loads(arg).items():
                    if 'kernel' in k.lower():
                        cm = ("#8dd3c7", "#fb8072", "#ffffb3", "#bebada", "#80b1d3",
                              "#fdb462", "#b3de69", "#fccde5")
                        summary['fillcolor'] = cm[v]
                    k = k[:1].upper() if len(k) > 4 else k
                    name += '{}{}.'.format(k, v)
                name += ')'
                summary['label'] = name
                return summary

            @staticmethod
            def get_config_grid(dict_space):
                param_grid = {}
                constants = {}
                for k, v in dict_space.items():
                    if isinstance(v, Categorical):
                        param_grid[k] = v.data
                    elif isinstance(v, Space):
                        raise NotImplementedError
                    else:
                        constants[k] = v
                from sklearn.model_selection import ParameterGrid
                configs = list(ParameterGrid(param_grid))
                for config in configs:
                    config.update(constants)
                return configs

        return enas_unit
    return registered_class

def enas_net(**kwvars):
    def registered_class(Cls):
        class ENAS_Net(Cls):
            def __init__(self, *args, **kwargs):
                kwvars.update(kwargs)
                super().__init__(*args, **kwvars)
                #
                self._modules = {}
                self._kwspaces = collections.OrderedDict()
                for k, module in kwvars.items():
                    if isinstance(module, (ENAS_Unit, ENAS_Sequential)):
                        self._modules[k] = module
                        if isinstance(module, ENAS_Unit):
                            self._kwspaces[k] = module.kwspaces
                        else:
                            assert isinstance(module, ENAS_Sequential)
                            for key, v in module.kwspaces.items():
                                new_key = '{}.{}'.format(k, key)
                                self._kwspaces[new_key] = v
                self.latency_evaluated = False
                self._avg_latency = 1

            @property
            def nparams(self):
                nparams = 0
                for k, op in self._modules.items():
                    if isinstance(op, (ENAS_Unit, ENAS_Sequential)):
                        nparams += op.nparams
                    else:
                        # standard block
                        for _, v in op.collect_params().items():
                            nparams += v.data().size
                return nparams

            @property
            def nodeend(self):
                return list(self._modules.keys())[-1]

            @property
            def nodehead(self):
                return list(self._modules.keys())[0]

            @property
            def graph(self):
                from graphviz import Digraph
                e = Digraph(node_attr={'color': 'lightblue2', 'style': 'filled', 'shape': 'box'})
                pre_node = 'input'
                e.node(pre_node)
                for k, op in self._modules.items():
                    if hasattr(op, 'graph'):
                        e.subgraph(op.graph)
                        e.edge(pre_node, op.nodehead)
                        pre_node = op.nodeend
                    else:
                        if hasattr(op, 'node'):
                            if op.node is None: continue
                            node_info = op.node
                        else:
                            node_info = {'label': op.__class__.__name__}
                        e.node(k, **node_info)
                        e.edge(pre_node, k)
                        pre_node = k
                return e


            @property
            def kwspaces(self):
                return self._kwspaces

            def sample(self, **configs):
                striped_keys = [k.split('.')[0] for k in configs.keys()]
                for k in striped_keys:
                    if isinstance(self._modules[k], ENAS_Unit):
                        self._modules[k].sample(configs[k])
                    else:
                        sub_configs = _strip_config_space(configs, prefix=k)
                        self._modules[k].sample(**sub_configs)

            @property
            def latency(self):
                if not self.latency_evaluated:
                    raise Exception('Latency is not evaluated yet.')
                return self._avg_latency

            @property
            def avg_latency(self):
                if not self.latency_evaluated:
                    raise Exception('Latency is not evaluated yet.')
                return self._avg_latency

            def evaluate_latency(self, x):
                import time
                # evaluate submodule latency
                for k, op in self._modules.items():
                    if hasattr(op, 'evaluate_latency'):
                        x = op.evaluate_latency(x)
                # calc avg_latency
                avg_latency = 0.0
                for k, op in self._modules.items():
                    if hasattr(op, 'avg_latency'):
                        avg_latency += op.avg_latency
                self._avg_latency = avg_latency
                self.latency_evaluated = True

        return ENAS_Net
    return registered_class

class ENAS_Sequential(gluon.HybridBlock):
    def __init__(self, *modules_list):
        """
        Args:
            modules_list(list of ENAS_Unit)
        """
        super().__init__()
        if len(modules_list) == 1 and isinstance(modules_list, (list, tuple)):
            modules_list = modules_list[0]
        self._modules = {}
        self._blocks = gluon.nn.HybridSequential()
        self._kwspaces = collections.OrderedDict()
        for i, op in enumerate(modules_list):
            self._modules[str(i)] = op
            with self._blocks.name_scope():
                self._blocks.add(op)
            if hasattr(op, 'kwspaces'):
                self._kwspaces[str(i)] = op.kwspaces
        self.latency_evaluated = False
        self._avg_latency = 1

    def __getitem__(self, index):
        return self._blocks[index]

    def hybrid_forward(self, F, x):
        for k, op in self._modules.items():
            x = op(x)
        return x

    def prune(self):
        _modules_keys = self._modules.keys()
        module_list = []
        for k in _modules_keys:
            if isinstance(self._modules[k], ENAS_Unit):
                index = self._modules[k].index
                op = self._modules[k].module_list[index]
                module_list.append(op)
            else:
                module_list.append(self._modules[k])
        return module_list

    @property
    def nodeend(self):
        return list(self._modules.keys())[-1]

    @property
    def nodehead(self):
        return list(self._modules.keys())[0]

    @property
    def graph(self):
        from graphviz import Digraph
        e = Digraph(node_attr={'color': 'lightblue2', 'style': 'filled', 'shape': 'box'})
        pre_node = None
        for i, op in self._modules.items():
            if hasattr(op, 'graph'):
                e.subgraph(op.graph)
                if pre_node:
                    e.edge(pre_node, op.nodehead)
                pre_node = op.nodeend
            else:
                if hasattr(op, 'node'):
                    if op.node is None: continue
                    node_info = op.node
                else:
                    node_info = {'label': op.__class__.__name__}
                e.node(i, **node_info)
                if pre_node:
                    e.edge(pre_node, i)
                pre_node = i
        return e

    @property
    def kwspaces(self):
        return self._kwspaces

    @property
    def nparams(self):
        nparams = 0
        for k, op in self._modules.items():
            if isinstance(op, ENAS_Unit):
                nparams += op.nparams
            else:
                # standard block
                for _, v in op.collect_params().items():
                    nparams += v.data().size
        return nparams

    @property
    def latency(self):
        if not self.latency_evaluated:
            raise Exception('Latency is not evaluated yet.')
        latency = 0.0
        for k, op in self._modules.items():
            if hasattr(op, 'latency'):
                latency += op.latency
        return latency

    @property
    def avg_latency(self):
        if not self.latency_evaluated:
            raise Exception('Latency is not evaluated yet.')
        return self._avg_latency

    def evaluate_latency(self, x):
        import time
        # evaluate submodule latency
        for k, op in self._modules.items():
            if hasattr(op, 'evaluate_latency'):
                x = op.evaluate_latency(x)
            else:
                x = op(x)
        # calc avg_latency
        avg_latency = 0.0
        for k, op in self._modules.items():
            if hasattr(op, 'avg_latency'):
                avg_latency += op.avg_latency
        self._avg_latency = avg_latency
        self.latency_evaluated = True
        return x

    def sample(self, **configs):
        for k, v in configs.items():
            self._modules[k].sample(v)

    def __repr__(self):
        reprstr = self.__class__.__name__ + '('
        for i, op in self._modules.items():
            reprstr += '\n\t{}: {}'.format(i, op)
        reprstr += ')\n'
        return reprstr

    def export(self, path):
        """Export HybridBlock to json format that can be loaded by
        `gluon.SymbolBlock.imports`, `mxnet.mod.Module` or the C++ interface.

        .. note:: When there are only one input, it will have name `data`. When there
                  Are more than one inputs, they will be named as `data0`, `data1`, etc.

        Parameters
        ----------
        path : str
            Path to save model. Two files `path-symbol.json` and `path.params`
            will be created.

        Examples
        --------

        >>> mynet.export('enas')
        >>> mynet_static = mx.gluon.nn.SymbolBlock.imports(
            "enas-symbol.json", ['data'], "enas.params")
        >>> x = mx.nd.random.uniform(shape=(1, 1, 28, 28))
        >>> y = mynet_static(x)
        """
        from mxnet import npx as _mx_npx
        from mxnet.util import is_np_array
        from mxnet import ndarray
        if not self._cached_graph:
            raise RuntimeError(
                "Please first call block.hybridize() and then run forward with "
                "this block at least once before calling export.")
        sym = self._cached_graph[1]
        sym.save('%s-symbol.json'%path, remove_amp_cast=True)

        arg_names = set(sym.list_arguments())
        aux_names = set(sym.list_auxiliary_states())
        arg_dict = {}
        for name, param in self.collect_params().items():
            if name in arg_names:
                arg_dict['arg:%s'%name] = param._reduce()
            elif name in aux_names:
                arg_dict['aux:%s'%name] = param._reduce()
            else:
                pass
        save_fn = _mx_npx.save if is_np_array() else ndarray.save
        save_fn('%s.params'%(path), arg_dict)

class Zero_Unit(gluon.HybridBlock):
    def hybrid_forward(self, F, x):
        return x
    def __repr__(self):
        return self.__class__.__name__

class ENAS_Unit(gluon.HybridBlock):
    def __init__(self, *ops, with_zero=False):
        super().__init__()
        self.module_list = gluon.nn.HybridSequential()
        self._latency = []
        for op in ops:
            self.module_list.add(op)
            self._latency.append(1)
        if with_zero:
            self.module_list.add(Zero_Unit())
            self._latency.append(1)
        self.index = 0
        self._latency_benchmark_times = 10
        self._latency_warmup_times = 5
        self.latency_evaluated = False

    def hybrid_forward(self, F, x):
        return self.module_list[self.index](x)

    @property
    def kwspaces(self):
        return Categorical(*list(range(len(self.module_list))))

    @property
    def nparams(self):
        nparams = 0
        for _, v in self.module_list[self.index].collect_params().items():
            nparams += v.data().size
        return nparams

    @property
    def latency(self):
        if not self.latency_evaluated:
            raise Exception('Latency is not evaluated yet.')
        return self._latency[self.index]

    @property
    def avg_latency(self):
        if not self.latency_evaluated:
            raise Exception('Latency is not evaluated yet.')
        return sum(self._latency) / len(self._latency)

    def evaluate_latency(self, x):
        import time
        for i, op in enumerate(self.module_list):
            latency_i = 0
            for j in range(self._latency_benchmark_times + self._latency_warmup_times):
                start_time = time.time() * 1000 # ms
                #print('op {}, shape x {}'.format(op, x.shape))
                y = op(x)
                mx.nd.waitall()
                end_time = time.time() * 1000 # ms
                if j > self._latency_warmup_times:
                    latency_i += end_time - start_time
            self._latency[i] = latency_i / self._latency_benchmark_times
        self.latency_evaluated = True
        return y

    def sample(self, ind):
        self.index = ind

    def __len__(self):
        return len(self.module_list)

    def __repr__(self):
        reprstr = self.__class__.__name__ + '(num of choices: {}), current architecture:\n\t {}' \
            .format(len(self.module_list), self.module_list[self.index])
        return reprstr



import logging
import os
import time
from typing import Any

from fastai.basic_train import Learner
from fastai.callbacks import EarlyStoppingCallback, TrackerCallback
from torch import Tensor

logger = logging.getLogger(__name__)


class EarlyStoppingCallbackWithTimeLimit(EarlyStoppingCallback):

    def __init__(self, learn: Learner, time_limit=None, **kwargs):
        super().__init__(learn, **kwargs)
        self.time_limit = time_limit
        self.start_time = time.time()

    def on_epoch_end(self, epoch, **kwargs):
        if self.time_limit:
            time_elapsed = time.time() - self.start_time
            time_left = self.time_limit - time_elapsed
            if time_left <= 0:
                logger.log(20, "\tRan out of time, stopping training early.")
                return {'stop_training': True}
        return super().on_epoch_end(epoch, **kwargs)


class SaveModelCallback(TrackerCallback):
    """A `TrackerCallback` that saves the model when monitored quantity is best."""

    def __init__(self, learn: Learner, monitor: str = 'valid_loss', mode: str = 'auto', every: str = 'improvement', name: str = 'bestmodel'):
        super().__init__(learn, monitor=monitor, mode=mode)
        self.every, self.name, self.best = every, name, None
        if self.every not in ['improvement', 'epoch']:
            logger.warning(f'SaveModel every {self.every} is invalid, falling back to "improvement".')
            self.every = 'improvement'

    def jump_to_epoch(self, epoch: int) -> None:
        try:
            self.learn.load(f'{self.name}_{epoch - 1}', purge=False)
            logger.info(15, f"Loaded {self.name}_{epoch - 1}")
        except:
            logger.info(15, f'Model {self.name}_{epoch - 1} not found.')

    def on_epoch_end(self, epoch: int, **kwargs: Any) -> None:
        """Compare the value monitored to its best score and maybe save the model."""
        if self.every == "epoch":
            self.learn.save(f'{self.name}_{epoch}')
        else:  # every="improvement"
            current = self.get_monitor_value()
            if isinstance(current, Tensor): current = current.cpu()
            if current is not None and self.operator(current, self.best):
                logger.log(15, f'Better model found at epoch {epoch} with {self.monitor} value: {current}.')
                self.best = current
                self.learn.save(f'{self.name}')

    def on_train_end(self, **kwargs):
        if self.every == "improvement" and os.path.isfile(self.path / self.model_dir / f'{self.name}.pth'):
            self.learn.load(f'{self.name}', purge=False)



import abc
import pandas as pd
import numpy as np
import json
import collections
from typing import List, Optional, Union, Tuple, Hashable
from . import constants as _C
from autogluon_contrib_nlp.base import INT_TYPES, BOOL_TYPES
from autogluon_contrib_nlp.data.vocab import Vocab

__all__ = ['CategoricalColumnProperty',
           'TextColumnProperty',
           'NumericalColumnProperty',
           'EntityColumnProperty']


class ColumnProperty(abc.ABC):
    """Column property means the general property of the columns.

    It performs basic featurization

    """
    type = _C.NULL

    def __init__(self, **kwargs):
        self._parsed = False
        self._freeze = False
        self._num_sample = None
        self._num_missing_samples = None
        self._name = None

    @property
    def name(self):
        return self._name

    @property
    def num_sample(self):
        return self._num_sample

    @property
    def num_missing_sample(self):
        return self._num_missing_samples

    @property
    def num_valid_sample(self):
        return self.num_sample - self.num_missing_sample

    def transform(self, ele):
        return ele

    @abc.abstractmethod
    def parse(self, column_data: pd.Series):
        """Parse the column data and fill in the necessary properties"""
        assert not self._parsed, 'Cannot call parse twice. ' \
                                 'Use "col_prop.clone()" and parse again.'
        self._parsed = True
        assert isinstance(column_data, pd.Series), \
            'Currently, the input column data must be a Pandas Series.'
        self._num_sample = len(column_data)
        self._num_missing_samples = column_data.isnull().sum().sum().item()
        self._name = column_data.name

    @abc.abstractmethod
    def clone(self):
        pass

    def info(self, additional_attributes=None):
        basename = self.__class__.__name__
        if basename.endswith('ColumnProperty'):
            basename = basename[:(-len('ColumnProperty'))]
        padding = 3
        ret = '{}(\n'.format(basename)
        ret += ' ' * padding + 'name="{}"\n'.format(self.name)
        ret += ' ' * padding + '#total/missing={}/{}\n'.format(self.num_sample,
                                                               self.num_missing_sample)
        if additional_attributes is not None:
            for key, info in additional_attributes:
                ret += ' ' * padding + '{}={}\n'.format(key, str(info))
        ret += ')\n'
        return ret

    @abc.abstractmethod
    def get_attributes(self):
        pass

    def __str__(self):
        return self.info()


class CategoricalColumnProperty(ColumnProperty):
    type = _C.CATEGORICAL

    def __init__(self,
                 categories: Optional[List[Union[int, bool, str]]] = None,
                 allow_missing: Optional[bool] = None):
        """

        Parameters
        ----------
        categories
            The possible categories
        allow_missing
            Whether the categorical column is allowed to contain missing values
        """
        super().__init__()
        self._allow_missing = allow_missing
        self._freq = None
        if categories is not None:
            if type(categories[0]).__module__ == np.__name__:
                categories = [ele.item() for ele in categories]
            assert allow_missing is not None
            if allow_missing:
                self._vocab = Vocab(categories)
            else:
                self._vocab = Vocab(categories, unk_token=None)
        else:
            self._vocab = None

    def transform(self, data: Hashable) -> int:
        """Transform the input data

        Parameters
        ----------
        data
            Element in the input data

        Returns
        -------
        idx
            The transformed idx
        """
        return self.to_idx(data)

    def inv_transform(self, idx: int) -> Hashable:
        """Transform the idx back to the category

        Parameters
        ----------
        idx

        Returns
        -------
        category
        """
        return self.to_category(idx)

    @property
    def num_class(self):
        return len(self._vocab)

    @property
    def num_non_special_class(self):
        return len(self._vocab.non_special_tokens)

    def to_idx(self, item):
        return self._vocab[item]

    def to_category(self, idx):
        return self._vocab.all_tokens[idx]

    @property
    def categories(self):
        if self._vocab is None:
            return None
        else:
            return self._vocab.non_special_tokens

    @property
    def frequencies(self):
        return self._freq

    @property
    def allow_missing(self):
        return self._allow_missing

    def parse(self, column_data: pd.Series):
        super().parse(column_data=column_data)
        if self._allow_missing is None:
            if self.num_missing_sample > 0:
                self._allow_missing = True
            else:
                self._allow_missing = False
        value_counts = column_data.value_counts()
        if self._vocab is None:
            categories = sorted(list(value_counts.keys()))
            if type(categories[0]).__module__ == np.__name__:
                categories = [ele.item() for ele in categories]
            if self._allow_missing:
                self._vocab = Vocab(tokens=categories)
            else:
                self._vocab = Vocab(tokens=categories, unk_token=None)
        self._freq = [value_counts[ele] if ele in value_counts else 0 for ele in self.categories]

    def clone(self):
        return CategoricalColumnProperty(categories=self.categories,
                                         allow_missing=self.allow_missing)

    def get_attributes(self):
        return {'categories': self.categories,
                'allow_missing': self.allow_missing}

    def info(self):
        return super().info(
            [('num_class (total/non_special)', '{}/{}'.format(self.num_class,
                                                              self.num_non_special_class)),
             ('categories', self.categories),
             ('freq', self.frequencies)])


class NumericalColumnProperty(ColumnProperty):
    type = _C.NUMERICAL

    def __init__(self, shape: Optional[Tuple] = None):
        """

        Parameters
        ----------
        shape
            The shape of the numerical values
        """
        super().__init__()
        self._shape = shape

    def transform(self, ele):
        if ele is None:
            return np.full(shape=self.shape, fill_value=np.nan, dtype=np.float32)
        else:
            return np.array(ele, dtype=np.float32)

    @property
    def shape(self):
        return self._shape

    def parse(self, column_data: pd.Series):
        super().parse(column_data)
        idx = column_data.first_valid_index()
        val = column_data[idx]
        inferred_shape = np.array(val).shape
        if self._shape is not None:
            assert self._shape == inferred_shape, 'Shape mismatch!. Expected shape={},' \
                                     ' shape in the dataset is {}'.format(self._shape,
                                                                          inferred_shape)
        else:
            self._shape = inferred_shape

    def clone(self):
        return NumericalColumnProperty(shape=self.shape)

    def get_attributes(self):
        return {'shape': self.shape}

    def info(self):
        return super().info([('shape', self.shape)])


class TextColumnProperty(ColumnProperty):
    type = _C.TEXT

    def __init__(self):
        super().__init__()
        self._min_length = None
        self._max_length = None
        self._avg_length = None

    @property
    def min_length(self):
        return self._min_length

    @property
    def max_length(self):
        return self._max_length

    @property
    def avg_length(self):
        return self._avg_length

    def parse(self, column_data: pd.Series):
        super().parse(column_data)
        lengths = column_data.apply(len)
        self._min_length = lengths.min()
        self._avg_length = lengths.mean()
        self._max_length = lengths.max()

    def clone(self):
        return TextColumnProperty()

    def get_attributes(self):
        return {}

    def info(self):
        return super().info([('length, min/avg/max',
                              '{:d}/{:.2f}/{:d}'.format(self.min_length,
                                                        self.avg_length, self.max_length))])


def _get_entity_label_type(label) -> str:
    """

    Parameters
    ----------
    label
        The label of an entity

    Returns
    -------
    type_str
        The type of the label. Will either be null, categorical or numerical
    """
    if label is None:
        return _C.NULL
    if isinstance(label, (int, str)):
        return _C.CATEGORICAL
    else:
        return _C.NUMERICAL


class EntityColumnProperty(ColumnProperty):
    """The Entities Column.

    The elements inside the column can be
    - a single dictionary -> 1 entity
    - a single tuple -> 1 entity
    - a list of dictionary -> K entities
    - a list of tuples -> K entities
    - an empty list -> 0 entity
    - None -> 0 entity

    For each entity, it will be
    1) a dictionary that contains these keys
    - start
        The character-level start of the entity
    - end
        The character-level end of the entity
    - label
        The label information of this entity.
        We support
        - categorical labels
            Each label can be either a unicode string or a int value.
        - numpy array/vector labels/numerical labels
            Each label should be a fixed-dimensional array/numerical value
    2) a tuple with (start, end) or (start, end, label)

    """
    type = _C.ENTITY

    def __init__(self, parent,
                 label_type=None,
                 label_shape=None,
                 label_keys=None):
        """

        Parameters
        ----------
        parent
            The column name of its parent
        label_type
            The type of the labels.
            Can be the following:
            - null
            - categorical
            - numerical
        label_shape
            The shape of the label. Only be available when the entity contains numerical label
        label_keys
            The vocabulary of the categorical label.
            It is only available when the entity contains categorical label.
        """
        super().__init__()
        self._parent = parent
        self._label_type = label_type
        self._label_shape = label_shape
        if self._label_shape is not None:
            self._label_shape = tuple(self._label_shape)
        if label_keys is not None:
            self._label_vocab = Vocab(tokens=label_keys,
                                      unk_token=None)
        else:
            self._label_vocab = None
        self._label_freq = None
        self._num_total_entity = None
        self._avg_entity_per_sample = None
        self._avg_span_length = None

    def transform(self,
                  data: Optional[Union[dict, List[dict],
                                       Tuple, List[Tuple]]]) -> Tuple[np.ndarray,
                                                                      Optional[np.ndarray]]:
        """Transform the element to a formalized format

        Returns
        -------
        char_offsets
            Numpy array. Shape is (#entities, 2)
        labels
            Either None, or the transformed label
            - None
                None
            - Categorical:
                (#entities,)
            - Numerical:
                (#entities,) + label_shape
        """
        if data is None:
            if self.label_type == _C.CATEGORICAL:
                return np.zeros((0, 2), dtype=np.int32),\
                       np.zeros((0,), dtype=np.int32)
            elif self.label_type == _C.NUMERICAL:
                return np.zeros((0, 2), dtype=np.int32), \
                       np.zeros((0,) + self.label_shape, dtype=np.float32)
            elif self.label_type == _C.NULL:
                return np.zeros((0, 2), dtype=np.int32), None
            else:
                raise NotImplementedError
        labels = None if self.label_type == _C.NULL else []
        char_offsets = []
        if isinstance(data, dict) or isinstance(data, tuple):
            data = [data]
        for ele in data:
            if isinstance(ele, dict):
                start = ele['start']
                end = ele['end']
                if self.label_type == _C.CATEGORICAL:
                    labels.append(self.label_to_idx(ele['label']))
                elif self.label_type == _C.NUMERICAL:
                    labels.append(ele['label'])
            else:
                start = ele[0]
                end = ele[1]
                if self.label_type == _C.CATEGORICAL:
                    labels.append(self.label_to_idx(ele[2]))
                elif self.label_type == _C.NUMERICAL:
                    labels.append(ele[2])
            char_offsets.append((start, end))
        char_offsets = np.stack(char_offsets)
        if self.label_type != _C.NULL:
            labels = np.stack(labels)
        return char_offsets, labels

    @property
    def label_shape(self) -> Optional[Tuple[int]]:
        """The shape of each individual label of the entity.

        Will only be enabled when label_type == numerical

        Returns
        -------
        ret
            The label shape
        """
        return self._label_shape

    @property
    def label_type(self) -> str:
        """Type of the label.

        If there is no label attached to the entities, it will return None.

        Returns
        -------
        ret
            The type of the label. Should be either
            - 'null'
            - 'categorical'
            - 'numerical'
        """
        return self._label_type

    def label_to_idx(self, label):
        assert self.label_type == _C.CATEGORICAL
        return self._label_vocab[label]

    def idx_to_label(self, idx):
        assert self.label_type == _C.CATEGORICAL
        return self._label_vocab.all_tokens[idx]

    @property
    def label_keys(self):
        if self._label_vocab is None:
            return None
        else:
            return self._label_vocab.non_special_tokens

    @property
    def label_freq(self):
        return self._label_freq

    @property
    def has_label(self):
        return self._label_type != _C.NULL

    @property
    def parent(self):
        return self._parent

    @property
    def avg_entity_per_sample(self):
        return self._avg_entity_per_sample

    @property
    def avg_span_length(self):
        return self._avg_span_length

    @property
    def num_total_entity(self):
        return self._num_total_entity

    def clone(self):
        return EntityColumnProperty(parent=self.parent,
                                    label_type=self.label_type,
                                    label_shape=self.label_shape,
                                    label_keys=self.label_keys)

    def get_attributes(self):
        return {'parent': self.parent,
                'label_type': self.label_type,
                'label_shape': self.label_shape,
                'label_keys': self.label_keys}

    def parse(self, column_data: pd.Series):
        super().parse(column_data)
        # Store statistics
        all_span_lengths = []
        categorical_label_counter = collections.Counter()
        for idx, entities in column_data.items():
            if entities is None:
                continue
            if isinstance(entities, dict) or isinstance(entities, tuple):
                entities = [entities]
            assert isinstance(entities, list),\
                'The entity type is "{}" and is not supported by ' \
                'GluonNLP. Received entities={}'.format(type(entities), entities)
            for entity in entities:
                if isinstance(entity, dict):
                    start = entity['start']
                    end = entity['end']
                    label = entity.get('label', None)
                else:
                    assert isinstance(entity, tuple)
                    if len(entity) == 2:
                        start, end = entity
                        label = None
                    else:
                        start, end, label = entity
                all_span_lengths.append(end - start)
                label_type = _get_entity_label_type(label)
                if label_type == _C.CATEGORICAL:
                    categorical_label_counter[label] += 1
                elif label_type == _C.NUMERICAL and self._label_shape is None:
                    self._label_shape = np.array(label).shape
                if self._label_type is not None:
                    assert self._label_type == label_type, \
                        'Unmatched label types. ' \
                        'The type of labels of all entities should be consistent. ' \
                        'Received label type="{}".' \
                        ' Stored label_type="{}"'.format(label_type, self._label_type)
                else:
                    self._label_type = label_type
        self._num_total_entity = len(all_span_lengths)
        self._avg_entity_per_sample = len(all_span_lengths) / self.num_valid_sample
        self._avg_span_length = np.mean(all_span_lengths).item()
        if self._label_type == _C.CATEGORICAL:
            if self._label_vocab is None:
                keys = sorted(categorical_label_counter.keys())
                self._label_vocab = Vocab(tokens=keys,
                                          unk_token=None)
                self._label_freq = [categorical_label_counter[ele] for ele in keys]
            else:
                for key in categorical_label_counter.keys():
                    if key not in self._label_vocab:
                        raise ValueError('The entity label="{}" is not found in the provided '
                                         'vocabulary. The provided labels="{}"'
                                         .format(key,
                                                 self._label_vocab.all_tokens))
                self._label_freq = [categorical_label_counter[ele]
                                    for ele in self._label_vocab.all_tokens]

    def info(self):
        additional_attributes = [('parent', '"{}"'.format(self._parent)),
                                 ('#total entity', self.num_total_entity),
                                 ('num entity per sample',
                                  '{:.2f}'.format(self.avg_entity_per_sample)),
                                 ('avg span length', '{:.2f}'.format(self._avg_span_length))]
        if self.label_type == _C.CATEGORICAL:
            additional_attributes.append(('num categories', len(self.label_keys)))
            additional_attributes.append(('max/min freq',
                                          '{}/{}'.format(max(self.label_freq),
                                                         min(self.label_freq))))
        elif self.label_type == _C.NUMERICAL:
            additional_attributes.append(('label_shape', self.label_shape))
        return super().info(additional_attributes)


def get_column_property_metadata(column_properties):
    metadata = dict()
    for col_name, col_prop in column_properties.items():
        metadata[col_name] = {'type': col_prop.type,
                              'attrs': col_prop.get_attributes()}
    return metadata


def get_column_properties_from_metadata(metadata):
    """Generate the column properties from metadata

    Parameters
    ----------
    metadata
        The path to the metadata json file. Or the loaded meta data

    Returns
    -------
    column_properties
        The column properties
    """
    column_properties = collections.OrderedDict()
    if metadata is None:
        return column_properties
    if isinstance(metadata, str):
        with open(metadata, 'r', encoding='utf-8') as f:
            metadata = json.load(f)
    else:
        assert isinstance(metadata, dict)
    for col_name in metadata:
        col_type = metadata[col_name]['type']
        col_attrs = metadata[col_name]['attrs']
        if col_type == _C.TEXT:
            column_properties[col_name] = TextColumnProperty(**col_attrs)
        elif col_type == _C.ENTITY:
            column_properties[col_name] = EntityColumnProperty(**col_attrs)
        elif col_type == _C.NUMERICAL:
            column_properties[col_name] = NumericalColumnProperty(**col_attrs)
        elif col_type == _C.CATEGORICAL:
            column_properties[col_name] = CategoricalColumnProperty(**col_attrs)
        else:
            raise KeyError('Column type is not supported.'
                           ' Type="{}"'.format(col_type))
    return column_properties



import os, logging

POINTER_SUFFIX = '.pointer'

logger = logging.getLogger(__name__)

# TODO: Add S3 support
def save(path, content_path, verbose=True):
    os.makedirs(os.path.dirname(path), exist_ok=True)

    f = open(path, "w")
    f.write(content_path)
    f.close()

    if verbose:
        logger.log(15, 'Saved pointer file to '+str(path)+' pointing to '+str(content_path))



from .searcher import BaseSearcher
from sklearn.model_selection import ParameterGrid

__all__ = ['GridSearcher']


class GridSearcher(BaseSearcher):
    """Grid Searcher that exhaustively tries all possible configurations. 
       This Searcher can only be used for discrete search spaces of type :class:`autogluon.space.Categorical`

    Examples
    --------
    >>> import autogluon as ag
    >>> @ag.args(
    ...     x=ag.space.Categorical(0, 1, 2),
    ...     y=ag.space.Categorical('a', 'b', 'c'))
    >>> def train_fn(args, reporter):
    ...     pass
    >>> searcher = ag.searcher.GridSearcher(train_fn.cs)
    >>> searcher.get_config()
    Number of configurations for grid search is 9
    {'x.choice': 2, 'y.choice': 2}
    """
    def __init__(self, configspace, **kwargs):
        super().__init__(
            configspace, reward_attribute=kwargs.get('reward_attribute'))
        param_grid = {}
        hp_ordering = configspace.get_hyperparameter_names()
        for hp in hp_ordering:
            hp_obj = configspace.get_hyperparameter(hp)
            hp_type = str(type(hp_obj)).lower()
            assert 'categorical' in hp_type, \
                'Only Categorical is supported, but {} is {}'.format(hp, hp_type)
            param_grid[hp] = hp_obj.choices

        self._configs = list(ParameterGrid(param_grid))
        print('Number of configurations for grid search is {}'.format(len(self._configs)))

    def __len__(self):
        return len(self._configs)

    def get_config(self):
        """ Return new hyperparameter configuration to try next.
        """
        return self._configs.pop()
